using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace PokeMod
{
	public class BreedChain
	{
		private static bool m_Initialized;
		private static List<PokemonInfo>[] m_MasterEggList = new List<PokemonInfo>[14];

		private BreedNode m_Node; //Initial Node
		private AttackInfo[] m_Attacks; //Just in case its needed

		private List<BreedNode> m_LatestChain;

		private List<BreedNode[]> m_SubNodes;
		public List<BreedNode[]> SubNodes{ get{ return m_SubNodes; } }
		public PokemonInfo Pokemon{ get { return m_Node.Pokemon; } }
		public BreedNode Node{ get{ return m_Node; } }

		public BreedChain( BreedNode node ) : this( node, new List<BreedNode>() )
		{
		}

		public BreedChain( BreedNode node, List<BreedNode> latest )
		{
			if ( !m_Initialized )
				Initialize();

			m_Node = node;
			m_Attacks = node.GetEggMoves(); //Only include remaining egg moves

			m_LatestChain = new List<BreedNode>();
			if ( latest != null )
				m_LatestChain.AddRange( latest );
			m_LatestChain.Add( node );
		}

		public BreedChain( PokemonInfo poke, params AttackInfo[] attacks ) : this( poke, new List<BreedNode>(), attacks )
		{
		}

		public BreedChain( PokemonInfo poke, List<BreedNode> latest, params AttackInfo[] attacks )
		{
			if ( !m_Initialized )
				Initialize();
			
			m_Attacks = attacks; //Initially we use every attack we are given
			m_Node = GetNode( poke );
			m_Attacks = m_Node.GetEggMoves(); //Refine the set to only include remaining egg moves

			m_LatestChain = new List<BreedNode>();
			if ( latest != null )
				m_LatestChain.AddRange( latest );
			m_LatestChain.Add( m_Node );
		}

		private static void Initialize()
		{
			List<PokemonInfo> masterlist = new List<PokemonInfo>();
			masterlist.AddRange( PokeData.GetPokemon() );
			masterlist.Remove( PokeData.FindPokemon( 132 ) );
			masterlist.Remove( PokeData.FindPokemon( 490 ) );
			masterlist.Remove( PokeData.FindPokemon( 489 ) );

			//Initialize each egg group 1 through 14 (not 0, or 15).
			m_MasterEggList = new List<PokemonInfo>[14];
			for ( int i = 0; i < m_MasterEggList.Length; i++ )
				m_MasterEggList[i] = new List<PokemonInfo>();

			for ( int i = 0; i < masterlist.Count; i++ )
			{
				PokemonInfo poke = masterlist[i];

				if ( !poke.FemaleOnly && !poke.Genderless && poke.EggGroup1 != EggGroup.NoEggs )
				{
						m_MasterEggList[(int)poke.EggGroup1 - 1].Add( poke );
						if ( poke.EggGroup2 != EggGroup.None )
							m_MasterEggList[(int)poke.EggGroup2 - 1].Add( poke );
				}
			}

			m_Initialized = true;
		}

		public void GetSubNodes()
		{
			if ( m_SubNodes == null )
				m_SubNodes = new List<BreedNode[]>();

			List<PokemonInfo> egglist = new List<PokemonInfo>();
			egglist.AddRange( m_MasterEggList[(int)Pokemon.EggGroup1 - 1] );

			if ( Pokemon.EggGroup2 != Pokemon.EggGroup1 && Pokemon.EggGroup2 != EggGroup.None )
			{
				List<PokemonInfo> egg2list = m_MasterEggList[(int)Pokemon.EggGroup2 - 1];

				for ( int i = 0;i < egg2list.Count; i++ )
					if ( !egglist.Contains( egg2list[i] ) )
						egglist.Add( egg2list[i] );
			}


			//Go through the final egg list
			GetSubNodesFromList( egglist );
		}

		private void GetSubNodesFromList( List<PokemonInfo> egglist )
		{
			for ( int i = 0; i < egglist.Count; i++ )
			{
				PokemonInfo poke = egglist[i];

				if ( poke != Pokemon && !poke.FemaleOnly && !poke.Genderless && poke.EggGroup1 != EggGroup.NoEggs )
				{
					bool first = poke.EggGroup1 == Pokemon.EggGroup1 || poke.EggGroup2 == Pokemon.EggGroup1;
					bool secnd = poke.EggGroup1 == Pokemon.EggGroup2 || poke.EggGroup2 == Pokemon.EggGroup2;

					if ( first || secnd )
					{
						//Get all of the moves (natural/egg) for that pokemon
						BreedNode node = GetNode( poke );

						if ( node.NatMoveCount > 0 || node.EggMoveCount > 0 )
							GetSubNodes( node );
					}
				}
			}
		}

		private void GetSubNodes( BreedNode node )
		{
			PokemonInfo poke = node.Pokemon;

			//Is this Pokemon already in our latest chain list?
			bool cantcont = false;
			int chaincnt = m_LatestChain.Count;

			for ( int j = 0;!cantcont && j < chaincnt; j++ )
			{
				if ( m_LatestChain[j].Pokemon == poke )
					cantcont = true;
			}

			//Pokemon from m_LatestChain[chaincnt-2] to first Pokemon with NatMoveCount > 0, if sharing any group

			//Can this Pokemon contribute to the chain?
			if ( chaincnt > 1 && m_Node.NatMoveCount == 0 && ( IsInBothGroups( poke, Pokemon ) || IsInAGroup( poke, m_LatestChain[chaincnt-2].Pokemon ) ) )
				cantcont = true;

			if ( m_LatestChain[chaincnt-1].NatMoveCount == 0 )
			{
				for ( int j = chaincnt-2;!cantcont && j >= 0; j-- )
				{
					if ( m_LatestChain[j].NatMoveCount == 0 )
					{ //Only has egg moves (intermediary or transition Pokemon)
						if ( IsInAGroup( poke, m_LatestChain[j].Pokemon ) ) //must be in at least one group
							cantcont = true;
					}
					else
						break;
				}
			}

			if ( cantcont )
				return;

			if ( node.NatMoveCount >= m_Attacks.Length )
				m_SubNodes.Add( new BreedNode[]{ node } );
			else if ( node.EggMoveCount > 0 && ( node.NatMoveCount > 0 || !IsInBothGroups( poke, Pokemon ) ) )
			{ //Has at least one natural attack, or transitions to another group
				//Get the subnode chains, node's Egg Moves will be the new attack set to look for

				BreedChain newchain = new BreedChain( node, m_LatestChain );
				newchain.GetSubNodes();
				List<BreedNode[]> subnodes = newchain.SubNodes;

				//Assimilate the breed chain's subnodes into the current chain's subnodes
				for ( int j = 0;j < subnodes.Count; j++ )
					m_SubNodes.Add( CombineNodes( subnodes[j], node ) );
			}
		}

		public static BreedNode[] CombineNodes( BreedNode[] chain, BreedNode node )
		{
			BreedNode[] newchain = new BreedNode[chain.Length+1];
			Array.Copy( chain, 0, newchain, 0, chain.Length );
			newchain[newchain.Length-1] = node;
			return newchain;
		}

		public static bool IsInAGroup( PokemonInfo first, PokemonInfo second )
		{
			EggGroup f1 = first.EggGroup1;
			EggGroup f2 = first.EggGroup2;
			EggGroup s1 = second.EggGroup1;
			EggGroup s2 = second.EggGroup2;

			return f1 == s1 || (s2 != EggGroup.None && f1 == s2) ||
				(f2 != EggGroup.None && f2 == s1) ||
				(s2 != EggGroup.None && f2 == s2);
		}

		public static bool IsInBothGroups( PokemonInfo first, PokemonInfo second )
		{
			EggGroup f1 = first.EggGroup1;
			EggGroup f2 = first.EggGroup2;
			EggGroup s1 = second.EggGroup1;
			EggGroup s2 = second.EggGroup2;

			return (f1 == s1 || (s2 != EggGroup.None && f1 == s2)) &&
				((f2 != EggGroup.None && f2 == s1) ||
				(s2 != EggGroup.None && f2 == s2));
		}

		private int GetAttackIndex( AttackInfo attack )
		{
			for ( int i = 0;i < m_Attacks.Length; i++ )
				if ( m_Attacks[i] == attack )
					return i;

			return -1;
		}

		private void AddMoveToNode( PokeMoveInfo[] moves, MoveType type, PokeGameFlag flag, ref BreedNode node )
		{
			for ( int j = 0;j < moves.Length; j++ )
			{
				int attind = GetAttackIndex( moves[j].Attack );

				if ( attind >= 0 )
					node.Add( moves[j], type, flag );
			}
		}

		private void GetEggMoves( BreedNode node )
		{
			GetEggMoves( node.Pokemon, node );
		}

		private void GetEggMoves( PokemonInfo info, BreedNode node )
		{
			AddMoveToNode( info.DPEggMoves, MoveType.Egg, PokeGameFlag.DP, ref node );

			AddMoveToNode( info.PtEggMoves, MoveType.Egg, PokeGameFlag.Platinum, ref node );

			AddMoveToNode( info.RSEggMoves, MoveType.Egg, PokeGameFlag.RuSa, ref node );

			AddMoveToNode( info.FLEggMoves, MoveType.Egg, PokeGameFlag.FrLg, ref node );

			AddMoveToNode( info.EEggMoves, MoveType.Egg, PokeGameFlag.Emerald, ref node );
		}

		private BreedNode GetNode( PokemonInfo info )
		{
			BreedNode node = new BreedNode( info );
			SetNodeAttacks( node );
	
			//Get attacks of every evolution, lets include baby
			PokemonInfo child = info;

			while ( child.EvoChild != null && !child.EvoChild.IsBaby )				
				SetNodeAttacks( (child = child.EvoChild), node );

			return node;
		}

		private void SetNodeAttacks( BreedNode node )
		{
			SetNodeAttacks( node.Pokemon, node );
		}

		private void SetNodeAttacks( PokemonInfo info, BreedNode node )
		{
			if ( info.KantoID == 235 ) //Smeargle
				for ( int i = 0; i < m_Attacks.Length; i++ )
					if ( m_Attacks[i].ID != 448 )
						node.Add( new PokeMoveInfo( m_Attacks[i].ID ), MoveType.Sketch, (PokeGameFlag)0x67 );

			for ( int i = 0;i < info.AltForms.Length; i++ )
			{
				AltFormInfo form = info.AltForms[i];

				AddMoveToNode( form.DPLevelUpMoves, MoveType.Level, PokeGameFlag.DP, ref node );

				AddMoveToNode( form.PtLevelUpMoves, MoveType.Level, PokeGameFlag.Platinum, ref node );

				AddMoveToNode( form.DPTMMoves, MoveType.TM, PokeGameFlag.DP, ref node );

				AddMoveToNode( form.PtTMMoves, MoveType.TM, PokeGameFlag.Platinum, ref node );

				AddMoveToNode( form.RSLevelUpMoves, MoveType.Level, PokeGameFlag.RuSa, ref node );

				AddMoveToNode( form.ELevelUpMoves, MoveType.Level, PokeGameFlag.Emerald, ref node );

				AddMoveToNode( form.FLLevelUpMoves, MoveType.Level, PokeGameFlag.FrLg, ref node );

				AddMoveToNode( form.RSTMMoves, MoveType.TM, PokeGameFlag.RuSa, ref node );

				AddMoveToNode( form.ETMMoves, MoveType.TM, PokeGameFlag.Emerald, ref node );

				AddMoveToNode( form.FLTMMoves, MoveType.TM, PokeGameFlag.FrLg, ref node );
			}

			AddMoveToNode( info.PtTutorMoves, MoveType.Tutor, PokeGameFlag.Platinum, ref node );

			AddMoveToNode( info.FLTutorMoves, MoveType.Tutor, PokeGameFlag.FrLg, ref node );

			AddMoveToNode( info.ETutorMoves, MoveType.Tutor, PokeGameFlag.Emerald, ref node );

			AddMoveToNode( info.XDTutorMoves, MoveType.Tutor, PokeGameFlag.XD, ref node );

			AddMoveToNode( info.DPTutorMoves, MoveType.Tutor, PokeGameFlag.DP, ref node );

			for ( int j = 0;j < info.PurifyAttacks.Length; j++ )
			{
				AttackInfo move = info.PurifyAttacks[j];
				int attind = GetAttackIndex( move );
				if ( attind > 0 )
					node.Add( new PokeMoveLevelInfo( move.ID, info.PurifyLevel ), MoveType.Purify, PokeGameFlag.XD );
			}

			GetEggMoves( info, node );
		}
	}

	public class PokeMoveEntry
	{
		private PokeMoveInfo m_Move;
		private MoveType m_Type;
		private PokeGameFlag m_Games;

		public PokeMoveInfo Move{ get{ return m_Move; } set{ m_Move = value; } }
		public MoveType Type{ get{ return m_Type; } set{ m_Type = value; } }
		public PokeGameFlag Games{ get{ return m_Games; } set{ m_Games = value; } }

		public PokeMoveEntry( PokeMoveInfo move, MoveType type, PokeGameFlag games )
		{
			m_Move = move;
			m_Type = type;
			m_Games = games;
		}

		public void AddGame( PokeGameFlag game )
		{
			m_Games |= game;
		}

		public bool IsEqual( PokeMoveInfo move, MoveType type )
		{
			return move == m_Move && type == m_Type;
		}
	}

	public class BreedNode
	{
		private PokemonInfo m_Pokemon;
		private List<PokeMoveEntry> m_NatMoves;
		private List<PokeMoveEntry> m_EggMoves;

		public PokemonInfo Pokemon{ get{ return m_Pokemon; } set{ m_Pokemon = value; } }
		public List<PokeMoveEntry> NatMoves{ get{ return m_NatMoves; } }
		public List<PokeMoveEntry> EggMoves{ get{ return m_EggMoves; } }
		public int NatMoveCount{ get{ return GetNatAttacks().Length; } }
		public int EggMoveCount{ get{ return GetEggMoves().Length; } }

		public BreedNode( PokemonInfo pokemon )
		{
			m_Pokemon = pokemon;
			m_NatMoves = new List<PokeMoveEntry>();
			m_EggMoves = new List<PokeMoveEntry>();
		}

		public void Add( PokeMoveInfo move, MoveType type, PokeGameFlag flag )
		{
			if ( type == MoveType.Egg )
			{
				//Can the Pokemon already do this naturally, somehow?
				for ( int i = 0;i < m_NatMoves.Count; i++ )
				{
					PokeMoveEntry entry = m_NatMoves[i];
					if ( entry.Move.Attack == move.Attack )
						return;
				}

				//Do we already have this in our list?  If so, add this game to it.
				for ( int i = 0;i < m_EggMoves.Count; i++ )
				{
					PokeMoveEntry entry = m_EggMoves[i];
					if ( entry.Move.Attack == move.Attack )
					{
						entry.AddGame( flag );
						return;
					}
				}

				//Otherwise, lets add it to our list.
				m_EggMoves.Add( new PokeMoveEntry( move, type, flag ) );
			}
			else
			{
				//Do we already have this in our list?  If so, add this game to it.
				for ( int i = 0;i < m_NatMoves.Count; i++ )
				{
					PokeMoveEntry entry = m_NatMoves[i];
					if ( entry.Move.Attack == move.Attack && type == entry.Type )
					{
						if ( type == MoveType.Level )
						{
							PokeMoveLevelInfo lvl1 = move as PokeMoveLevelInfo;
							PokeMoveLevelInfo lvl2 = m_NatMoves[i].Move as PokeMoveLevelInfo;
							if ( lvl1 != null && lvl2 != null && lvl1.Level == lvl2.Level )
							{
								entry.AddGame( flag );
								return;
							}
						}
						else
						{
							entry.AddGame( flag );
							return;
						}
					}
				}

				//Otherwise, lets add it to our list.
				m_NatMoves.Add( new PokeMoveEntry( move, type, flag ) );
			}
		}

		public AttackInfo[] GetEggMoves()
		{
			List<AttackInfo> attacks = new List<AttackInfo>();
			for ( int i = 0;i < m_EggMoves.Count; i++ )
				if ( !attacks.Contains( m_EggMoves[i].Move.Attack ) )
					attacks.Add( m_EggMoves[i].Move.Attack );

			return attacks.ToArray();
		}

		public AttackInfo[] GetNatAttacks()
		{
			List<AttackInfo> attacks = new List<AttackInfo>();
			for ( int i = 0;i < m_NatMoves.Count; i++ )
				if ( !attacks.Contains( m_NatMoves[i].Move.Attack ) )
					attacks.Add( m_NatMoves[i].Move.Attack );

			return attacks.ToArray();
		}

		public override string ToString()
		{
			return ToString( true );
		}

		public string ToString( bool dispegg )
		{
			return ToString( false, dispegg );
		}

		public string ToString( bool evo, bool dispegg )
		{
			StringBuilder builder = new StringBuilder();

			PokemonInfo poke = m_Pokemon;

			if ( evo )
			{
				while ( poke.EvoChild != null && ( poke.EvoChild.EvoChild != null || poke.EvoChain.BreedItem == null ) )
					poke = poke.EvoChild;

				builder.Append( poke.PokeNames[1].ToString() );

				if ( poke != m_Pokemon && poke.EggGroup1 == EggGroup.NoEggs ) //Baby pokes
				{
					builder.Append( " -> " );
					builder.Append( m_Pokemon.PokeNames[1].ToString() );
				}
			}
			else if ( poke.IsBaby )
			{

				List<PokemonInfo> pokelist = new List<PokemonInfo>();
				pokelist.AddRange( PokeData.GetPokemon() );

				for ( int i = 0; i < pokelist.Count; i++ )
				{
					if ( pokelist[i].EvoChild == poke )
					{
						poke = pokelist[i];
						break;
					}
				}

				builder.Append( m_Pokemon.PokeNames[1].ToString() );

				if ( poke != m_Pokemon )
				{
					builder.Append( " -> " );
					builder.Append( poke.PokeNames[1].ToString() );
				}

			}
			else
				builder.Append( m_Pokemon.PokeNames[1].ToString() );

/*
			if ( m_NatMoves.Count > 0 || ( dispegg && m_EggMoves.Count > 0 ) )
			{
				builder.Append( " - " );

				int comma = 0;

				for ( int i = 0;i < m_NatMoves.Count; i++ )
				{
					PokeMoveEntry entry = m_NatMoves[i];

					if ( comma > 0 )
						builder.Append( ", " );

					builder.Append( entry.Move.Attack.Name );
					builder.Append( " (" );
					string movetype = String.Empty;
					switch ( entry.Type )
					{
						case MoveType.Level: builder.Append( String.Format( "level:{0}", (entry.Move as PokeMoveLevelInfo).Level ) ); break;
						case MoveType.Tutor: builder.Append( "tutor" ); break;
						case MoveType.Purify: builder.Append( String.Format( "purify:{0}", m_Pokemon.PurifyLevel ) ); break;
						case MoveType.TM: builder.Append( "tm" ); break;
						case MoveType.Sketch: builder.Append( "sketch" ); break;
					}
					builder.Append( ":" );
					builder.Append( ((int)entry.Games).ToString("X") );
					builder.Append( ")" );
					comma++;
				}

				if ( dispegg )
					for ( int i = 0;i < m_EggMoves.Count; i++ )
					{
						PokeMoveEntry entry = m_EggMoves[i];

						if ( comma > 0 )
							builder.Append( ", " );

						builder.Append( entry.Move.Attack.Name );
						builder.Append( " (egg" );
						builder.Append( ":" );
						builder.Append( ((int)entry.Games).ToString("X") );
						builder.Append( ")" );
						comma++;
					}
			}
*/
			return builder.ToString();
		}
	}
}